@@ -122,7 +122,7 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 		}
 
 		$sqlString = 'INSERT INTO ' . $tableName . ' (' . implode(', ', $fields) . ') VALUES (' . implode(', ', $values) . ')';
-		$this->replacePlaceholders($sqlString, $parameters);
+		$this->replacePlaceholders($sqlString, $parameters, $tableName);
 		// debug($sqlString,-2);
 		$this->databaseHandle->sql_query($sqlString);
 		$this->checkSqlErrors($sqlString);
@@ -154,7 +154,7 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 		$parameters[] = $uid;
 
 		$sqlString = 'UPDATE ' . $tableName . ' SET ' . implode(', ', $fields) . ' WHERE uid=?';
-		$this->replacePlaceholders($sqlString, $parameters);
+		$this->replacePlaceholders($sqlString, $parameters, $tableName);
 		// debug($sqlString,-2);
 		$returnValue = $this->databaseHandle->sql_query($sqlString);
 		$this->checkSqlErrors($sqlString);
@@ -174,7 +174,7 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 	 */
 	public function removeRow($tableName, array $identifier, $isRelation = FALSE) {
 		$statement = 'DELETE FROM ' . $tableName . ' WHERE ' . $this->parseIdentifier($identifier);
-		$this->replacePlaceholders($statement, $identifier);
+		$this->replacePlaceholders($statement, $identifier, $tableName);
 		if (!$isRelation && isset($identifier['uid'])) {
 			$this->clearPageCache($tableName, $identifier['uid'], $isRelation);
 		}
@@ -193,7 +193,7 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 	 */
 	public function getRowByIdentifier($tableName, array $identifier) {
 		$statement = 'SELECT * FROM ' . $tableName . ' WHERE ' . $this->parseIdentifier($identifier);
-		$this->replacePlaceholders($statement, $identifier);
+		$this->replacePlaceholders($statement, $identifier, $tableName);
 		// debug($statement,-2);
 		$res = $this->databaseHandle->sql_query($statement);
 		$this->checkSqlErrors($statement);
@@ -233,7 +233,11 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 			$statementParts = $this->parseQuery($query, $parameters);
 			$sql = $this->buildQuery($statementParts, $parameters);
 		}
-		$this->replacePlaceholders($sql, $parameters);
+		$tableName = 'foo';
+		if (is_array($statementParts) && !empty($statementParts['tables'][0])) {
+			$tableName = $statementParts['tables'][0];
+		}
+		$this->replacePlaceholders($sql, $parameters, $tableName);
 		// debug($sql,-2);
 		$result = $this->databaseHandle->sql_query($sql);
 		$this->checkSqlErrors($sql);
@@ -260,14 +264,20 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 		// if limit is set, we need to count the rows "manually" as COUNT(*) ignores LIMIT constraints
 		if (!empty($statementParts['limit'])) {
 			$statement = $this->buildQuery($statementParts, $parameters);
-			$this->replacePlaceholders($statement, $parameters);
+			$this->replacePlaceholders($statement, $parameters, $statementParts['tables'][0]);
 			$result = $this->databaseHandle->sql_query($statement);
 			$this->checkSqlErrors($statement);
 			$count = $this->databaseHandle->sql_num_rows($result);
 		} else {
 			$statementParts['fields'] = array('COUNT(*)');
+			// having orderings without grouping is not compatible with non-MySQL DBMS
+			$statementParts['orderings'] = array();
+			if (isset($statementParts['keywords']['distinct'])) {
+				unset($statementParts['keywords']['distinct']);
+				$statementParts['fields'] = array('COUNT(DISTINCT ' . reset($statementParts['tables']) . '.uid)');
+			}
 			$statement = $this->buildQuery($statementParts, $parameters);
-			$this->replacePlaceholders($statement, $parameters);
+			$this->replacePlaceholders($statement, $parameters, $statementParts['tables'][0]);
 			$result = $this->databaseHandle->sql_query($statement);
 			$this->checkSqlErrors($statement);
 			$rows = $this->getRowsFromResult($query->getSource(), $result);
@@ -369,7 +379,7 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 		if (!empty($sql['additionalWhereClause'])) {
 			$statement .= ' AND ' . implode(' AND ', $sql['additionalWhereClause']);
 		}
-		$this->replacePlaceholders($statement, $parameters);
+		$this->replacePlaceholders($statement, $parameters, $tableName);
 		// debug($statement,-2);
 		$res = $this->databaseHandle->sql_query($statement);
 		$this->checkSqlErrors($statement);
@@ -402,7 +412,7 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 	}
 
 	/**
-	 * Adda a constrint to ensure that the record type of the returned tuples is matching the data type of the repository.
+	 * Add a constraint to ensure that the record type of the returned tuples is matching the data type of the repository.
 	 *
 	 * @param string $className The class name
 	 * @param array &$sql The query parts
@@ -425,7 +435,8 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 				if (count($recordTypes) > 0) {
 					$recordTypeStatements = array();
 					foreach ($recordTypes as $recordType) {
-						$recordTypeStatements[] = $dataMap->getTableName() . '.' . $dataMap->getRecordTypeColumnName() . '=' . $this->databaseHandle->fullQuoteStr($recordType, 'foo');
+						$tableName = $dataMap->getTableName();
+						$recordTypeStatements[] = $tableName . '.' . $dataMap->getRecordTypeColumnName() . '=' . $this->databaseHandle->fullQuoteStr($recordType, $tableName);
 					}
 					$sql['additionalWhereClause'][] = '(' . implode(' OR ', $recordTypeStatements) . ')';
 				}
@@ -558,15 +569,17 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 				$typeOfRelation = $columnMap->getTypeOfRelation();
 				if ($typeOfRelation === Tx_Extbase_Persistence_Mapper_ColumnMap::RELATION_HAS_AND_BELONGS_TO_MANY) {
 					$relationTableName = $columnMap->getRelationTableName();
-					$sql['where'][] = $tableName . '.uid IN (SELECT ' . $columnMap->getParentKeyFieldName() . ' FROM ' . $relationTableName . ' WHERE ' . $columnMap->getChildKeyFieldName() . '=' . $this->getPlainValue($operand2) . ')';
+					$sql['where'][] = $tableName . '.uid IN (SELECT ' . $columnMap->getParentKeyFieldName() . ' FROM ' . $relationTableName . ' WHERE ' . $columnMap->getChildKeyFieldName() . '=?)';
+					$parameters[] = intval($this->getPlainValue($operand2));
 				} elseif ($typeOfRelation === Tx_Extbase_Persistence_Mapper_ColumnMap::RELATION_HAS_MANY) {
 					$parentKeyFieldName = $columnMap->getParentKeyFieldName();
 					if (isset($parentKeyFieldName)) {
 						$childTableName = $columnMap->getChildTableName();
-						$sql['where'][] = $tableName . '.uid=(SELECT ' . $childTableName . '.' . $parentKeyFieldName . ' FROM ' . $childTableName . ' WHERE ' . $childTableName . '.uid=' . $this->getPlainValue($operand2) . ')';
+						$sql['where'][] = $tableName . '.uid=(SELECT ' . $childTableName . '.' . $parentKeyFieldName . ' FROM ' . $childTableName . ' WHERE ' . $childTableName . '.uid=?)';
+						$parameters[] = intval($this->getPlainValue($operand2));
 					} else {
-						$statement = 'FIND_IN_SET(' . $this->getPlainValue($operand2) . ',' . $tableName . '.' . $columnName . ')';
-						$sql['where'][] = $statement;
+						$sql['where'][] = 'FIND_IN_SET(?,' . $tableName . '.' . $columnName . ')';
+						$parameters[] = intval($this->getPlainValue($operand2));
 					}
 				} else {
 					throw new Tx_Extbase_Persistence_Exception_RepositoryException('Unsupported relation for contains().', 1267832524);
@@ -739,9 +752,11 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 	 *
 	 * @param string $sqlString The query part with placeholders
 	 * @param array $parameters The parameters
+	 * @param string $tableName
+	 *
 	 * @return string The query part with replaced placeholders
 	 */
-	protected function replacePlaceholders(&$sqlString, array $parameters) {
+	protected function replacePlaceholders(&$sqlString, array $parameters, $tableName = 'foo') {
 		// TODO profile this method again
 		if (substr_count($sqlString, '?') !== count($parameters)) throw new Tx_Extbase_Persistence_Exception('The number of question marks to replace must be equal to the number of parameters.', 1242816074);
 		$offset = 0;
@@ -753,11 +768,11 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 				} elseif (is_array($parameter) || ($parameter instanceof ArrayAccess) || ($parameter instanceof Traversable)) {
 					$items = array();
 					foreach ($parameter as $item) {
-						$items[] = $this->databaseHandle->fullQuoteStr($item, 'foo');
+						$items[] = $this->databaseHandle->fullQuoteStr($item, $tableName);
 					}
 					$parameter = '(' . implode(',', $items) . ')';
 				} else {
-					$parameter = $this->databaseHandle->fullQuoteStr($parameter, 'foo'); // FIXME This may not work with DBAL; check this
+					$parameter = $this->databaseHandle->fullQuoteStr($parameter, $tableName);
 				}
 				$sqlString = substr($sqlString, 0, $markPosition) . $parameter . substr($sqlString, $markPosition + 1);
 			}
@@ -774,16 +789,14 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 	 * @return void
 	 */
 	protected function addAdditionalWhereClause(Tx_Extbase_Persistence_QuerySettingsInterface $querySettings, $tableName, &$sql) {
-		if ($querySettings instanceof Tx_Extbase_Persistence_Typo3QuerySettings) {
-			if ($querySettings->getRespectEnableFields()) {
-				$this->addEnableFieldsStatement($tableName, $sql);
-			}
-			if ($querySettings->getRespectSysLanguage()) {
-				$this->addSysLanguageStatement($tableName, $sql);
-			}
-			if ($querySettings->getRespectStoragePage()) {
-				$this->addPageIdStatement($tableName, $sql, $querySettings->getStoragePageIds());
-			}
+		if ($querySettings->getRespectEnableFields()) {
+			$this->addEnableFieldsStatement($tableName, $sql);
+		}
+		if ($querySettings->getRespectSysLanguage()) {
+			$this->addSysLanguageStatement($tableName, $sql);
+		}
+		if ($querySettings->getRespectStoragePage()) {
+			$this->addPageIdStatement($tableName, $sql, $querySettings->getStoragePageIds());
 		}
 	}
 
@@ -837,6 +850,9 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 			$this->tableInformationCache[$tableName]['columnNames'] = $this->databaseHandle->admin_get_fields($tableName);
 		}
 		if (is_array($GLOBALS['TCA'][$tableName]['ctrl']) && array_key_exists('pid', $this->tableInformationCache[$tableName]['columnNames'])) {
+			if (empty($storagePageIds)) {
+				throw new Tx_Extbase_Persistence_Exception_InconsistentQuerySettings('Missing storage page ids.', 1365779762);
+			}
 			$sql['additionalWhereClause'][] = $tableName . '.pid IN (' . implode(', ', $storagePageIds) . ')';
 		}
 	}
@@ -893,9 +909,9 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 	 */
 	protected function parseLimitAndOffset($limit, $offset, array &$sql) {
 		if ($limit !== NULL && $offset !== NULL) {
-			$sql['limit'] = $offset . ', ' . $limit;
+			$sql['limit'] = intval($offset) . ', ' . intval($limit);
 		} elseif ($limit !== NULL) {
-			$sql['limit'] = $limit;
+			$sql['limit'] = intval($limit);
 		}
 	}
 
@@ -910,9 +926,6 @@ class Tx_Extbase_Persistence_Storage_Typo3DbBackend implements Tx_Extbase_Persis
 		$rows = array();
 		while ($row = $this->databaseHandle->sql_fetch_assoc($result)) {
 			if (is_array($row)) {
-				// TODO Check if this is necessary, maybe the last line is enough
-				$arrayKeys = range(0, count($row));
-				array_fill_keys($arrayKeys, $row);
 				$rows[] = $row;
 			}
 		}
